#ifndef __Viscosity_Star2021_h__
#define __Viscosity_Star2021_h__

#include "SPlisHSPlasH/Common.h"
#include "SPlisHSPlasH/FluidModel.h"
#include "ViscosityBase.h"
#include "SPlisHSPlasH/Utilities/MatrixFreeSolver.h"

#define USE_BLOCKDIAGONAL_PRECONDITIONER

namespace SPH
{
	/** \brief This class implements the implicit Laplace viscosity method introduced 
	* by Weiler et al. 2018 [WKBB18].
	*
	* References:
	* - [WKBB18] Marcel Weiler, Dan Koschier, Magnus Brand, and Jan Bender. A physically consistent implicit viscosity solver for SPH fluids. Computer Graphics Forum (Eurographics), 2018. URL: https://doi.org/10.1111/cgf.13349
    * 
    * Add by Starodubtsev
	*/
	class Viscosity_Star2021 : public ViscosityBase
	{
	protected:
		Real m_boundaryViscosity;
		unsigned int m_maxIter;
		Real m_maxError;
		unsigned int m_iterations;
		std::vector<Vector3r> m_vDiff;
		Real m_tangentialDistanceFactor;
        
        std::vector<Real> m_viscoCoeffs;
        std::vector<Real> m_lifeTime;
        std::vector<Real> m_spawnTime;
        Real m_viscoStep;

#ifdef USE_BLOCKDIAGONAL_PRECONDITIONER
		typedef Eigen::ConjugateGradient<MatrixReplacement, Eigen::Lower | Eigen::Upper, BlockJacobiPreconditioner3D> Solver;
		FORCE_INLINE static void diagonalMatrixElement(const unsigned int row, Matrix3r &result, void *userData);
#else
		typedef Eigen::ConjugateGradient<MatrixReplacement, Eigen::Lower | Eigen::Upper, JacobiPreconditioner3D> Solver;
		FORCE_INLINE static void diagonalMatrixElement(const unsigned int row, Vector3r &result, void *userData);
#endif	
		Solver m_solver;


		virtual void initParameters();
		
	public:
		static int ITERATIONS;
		static int MAX_ITERATIONS;
		static int MAX_ERROR;
		static int VISCOSITY_COEFFICIENT_BOUNDARY;
        static int VISCOSITY_STEP;
        
        static NonPressureForceBase* creator(FluidModel* model) { return new Viscosity_Star2021(model); }

		Viscosity_Star2021(FluidModel *model);
		virtual ~Viscosity_Star2021(void);

		virtual void step();
		virtual void reset();

		virtual void performNeighborhoodSearchSort();

		static void matrixVecProd(const Real* vec, Real *result, void *userData);

		FORCE_INLINE const Vector3r& getVDiff(const unsigned int i) const
		{
			return m_vDiff[i];
		}

		FORCE_INLINE Vector3r& getVDiff(const unsigned int i)
		{
			return m_vDiff[i];
		}

		FORCE_INLINE void setVDiff(const unsigned int i, const Vector3r& val)
		{
			m_vDiff[i] = val;
		}
		
		
		
		FORCE_INLINE const Real getViscoCoeffs(const unsigned int i) const
		{
			return m_viscoCoeffs[i];
		}

		FORCE_INLINE Real& getViscoCoeffs(const unsigned int i)
		{
			return m_viscoCoeffs[i];
		}

		FORCE_INLINE void setViscoCoeffs(const unsigned int i, const Real val)
		{
			m_viscoCoeffs[i] = val;
		}
		
		
		FORCE_INLINE const Real getLifeTime(const unsigned int i) const
		{
			return m_lifeTime[i];
		}

		FORCE_INLINE Real& getLifeTime(const unsigned int i)
		{
			return m_lifeTime[i];
		}

		FORCE_INLINE void setLifeTime(const unsigned int i, const Real val)
		{
			m_lifeTime[i] = val;
		}
	};
}

#endif
